/*
PAELLOS - Plateforme d'Apprentissage En Ligne Libre Orientée Sémantique

Copyright (C) 2011 Observatoire de Paris

Ce programme est un logiciel libre ; vous pouvez le redistribuer et/ou le modifier conformément aux dispositions de la Licence Publique Générale GNU, telle que publiée par la Free Software Foundation ; version 2 de la licence, ou encore (à votre choix) toute version ultérieure.

Ce programme est distribué dans l'espoir qu'il sera utile, mais SANS AUCUNE GARANTIE ; sans même la garantie implicite de COMMERCIALISATION ou D'ADAPTATION A UN OBJET PARTICULIER. Pour plus de détail, voir la Licence Publique Générale GNU .

Vous devez avoir reçu un exemplaire de la Licence Publique Générale GNU en même temps que ce programme ; si ce n'est pas le cas, écrivez à la Free Software Foundation Inc., 675 Mass Ave, Cambridge, MA 02139, Etats-Unis.
*/

package paellos;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.FontMetrics;
import java.awt.Graphics2D;
import java.awt.RenderingHints;
import java.awt.Transparency;
import java.awt.color.ColorSpace;
import java.awt.font.GlyphVector;
import java.awt.geom.Rectangle2D;
import java.awt.image.BufferedImage;
import java.awt.image.ColorModel;
import java.awt.image.ComponentColorModel;
import java.awt.image.DataBuffer;
import java.awt.image.WritableRaster;
import java.io.IOException;

import javax.imageio.IIOImage;
import javax.imageio.ImageIO;
import javax.imageio.ImageTypeSpecifier;
import javax.imageio.ImageWriter;
import javax.imageio.metadata.IIOInvalidTreeException;
import javax.imageio.metadata.IIOMetadata;
import javax.imageio.metadata.IIOMetadataNode;

import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import jaxe.elements.JESymbole2;


/**
 * Comme avec Equation, pour la création d'une image correspondant au texte.
 * Utilise la police STIXSubset-Regular.ttf
 *
 * @author Damien Guillaume
 */
class Symbole {
    
    /**
     * Création de l'image correspondant au texte du symbole (généralement un caractère, mais parfois 2 si c'est du utf32)
     */
    public static BufferedImage creerImage(final String texteSymbole) {
        final Font stix = JESymbole2.getSTIXFont();
        // on est obligé de créer un Graphics bidon pour calculer la taille de l'image depuis que la méthode
        // Toolkit.getFontMetrics(Font) est dépréciée
        final BufferedImage imgcalc = new BufferedImage(10, 10, BufferedImage.TYPE_INT_ARGB);
        final Graphics2D gcalc = imgcalc.createGraphics();
        gcalc.setFont(stix);
        final Graphics2D gcalc2d = (Graphics2D)gcalc;
        gcalc2d.setRenderingHint(RenderingHints.KEY_TEXT_ANTIALIASING, RenderingHints.VALUE_TEXT_ANTIALIAS_ON);
        final FontMetrics fm = gcalc.getFontMetrics(stix);
        final Rectangle2D r = fm.getStringBounds(texteSymbole, gcalc);
        final int l = (int)Math.ceil(r.getWidth());
        final int axs2 = getAscentHeight(gcalc2d, "x") / 2;
        final int asym = getAscentHeight(gcalc2d, texteSymbole);
        final int dsym = getDescentHeight(gcalc2d, texteSymbole);
        final int h = Math.max(asym - axs2, dsym + axs2) * 2;
        
        // IE ne gère pas les PNG avec une palette -> utilisation d'une échelle de gris avec transparence
        final ColorSpace grayColorSpace = ColorSpace.getInstance(ColorSpace.CS_GRAY);
        final ColorModel cm = new ComponentColorModel(grayColorSpace, true, true, Transparency.TRANSLUCENT, DataBuffer.TYPE_BYTE);
        final WritableRaster raster = cm.createCompatibleWritableRaster(l, h);
        final BufferedImage img2 = new BufferedImage(cm, raster, false, null);
        final Graphics2D g2 = img2.createGraphics();
        g2.setColor(Color.black);
        g2.setFont(stix);
        g2.setRenderingHint(RenderingHints.KEY_TEXT_ANTIALIASING, RenderingHints.VALUE_TEXT_ANTIALIAS_ON);
        g2.drawString(texteSymbole, 0, h/2 + axs2);
        g2.dispose();
        return img2;
    }
    
    private static int getAscentHeight(final Graphics2D g2d, final String s) {
        final GlyphVector gv = g2d.getFont().createGlyphVector(g2d.getFontRenderContext(), s.toCharArray());
        final Rectangle2D r = gv.getVisualBounds();
        double miny = r.getMinY();
        if (miny > 0)
            miny = 0;
        double maxy = r.getMaxY();
        if (maxy > 0)
            maxy = 0;
        return((int)Math.round(maxy - miny));
    }
    
    private static int getDescentHeight(final Graphics2D g2d, final String s) {
        final GlyphVector gv = g2d.getFont().createGlyphVector(g2d.getFontRenderContext(), s.toCharArray());
        final Rectangle2D r = gv.getVisualBounds();
        double miny = r.getMinY();
        if (miny < 0)
            miny = 0;
        double maxy = r.getMaxY();
        if (maxy < 0)
            maxy = 0;
        return((int)Math.round(maxy - miny));
    }
    
    /**
     * Enregistrement de l'image du symbole vers une sortie donnée en paramètre.
     * Le paramètre sortie peut être de la classe File (pour enregistrer dans un fichier) ou OutputStream.
     * Les images peuvent être utilisées dans IE6 (PNG avec une échelle de gris et fond blanc, couche alpha
     * utilisable par les autres navigateurs).
     */
    public static void enregistrerImage(final BufferedImage img, final Object sortie) throws IOException {
        // IE met un fond gris si aucun bKGD n'est spécifié -> ajout de bKGD dans les métadonnées
        final ImageWriter premierPourPNG = ImageIO.getImageWritersByFormatName("PNG").next();
        final IIOMetadata metadata = premierPourPNG.getDefaultImageMetadata(new ImageTypeSpecifier(img), null);
        final Element racine = (Element)metadata.getAsTree("javax_imageio_png_1.0");
        final NodeList nl = racine.getElementsByTagName("bKGD");
        if (nl.getLength() == 0) {
            Node suivant = racine.getFirstChild();
            final String name = suivant.getNodeName();
            if (("IHDR".equals(name) || "PLTE".equals(name)))
                suivant = suivant.getNextSibling();
            if (suivant != null && "PLTE".equals(suivant.getNodeName()))
                suivant = suivant.getNextSibling();
            final Element bKGD = new IIOMetadataNode("bKGD");
            if (suivant == null)
                racine.appendChild(bKGD);
            else
                racine.insertBefore(bKGD, suivant);
            final Element bKGD_Grayscale = new IIOMetadataNode("bKGD_Grayscale");
            bKGD.appendChild(bKGD_Grayscale);
            bKGD_Grayscale.setAttribute("gray", "255");
            try {
                metadata.mergeTree("javax_imageio_png_1.0", racine);
            } catch (final IIOInvalidTreeException ex) {
                System.err.println("enregistrerImage(BufferedImage, Object): " +  ex.getClass() + ": " + ex.getMessage());
            }
        }
        // enregistrement (peut lancer une IOException)
        premierPourPNG.setOutput(ImageIO.createImageOutputStream(sortie));
        premierPourPNG.write(new IIOImage(img, null, metadata));
    }
    
}
