/*
PAELLOS - Plateforme d'Apprentissage En Ligne Libre Orientée Sémantique

Copyright (C) 2011 Observatoire de Paris

Ce programme est un logiciel libre ; vous pouvez le redistribuer et/ou le modifier conformément aux dispositions de la Licence Publique Générale GNU, telle que publiée par la Free Software Foundation ; version 2 de la licence, ou encore (à votre choix) toute version ultérieure.

Ce programme est distribué dans l'espoir qu'il sera utile, mais SANS AUCUNE GARANTIE ; sans même la garantie implicite de COMMERCIALISATION ou D'ADAPTATION A UN OBJET PARTICULIER. Pour plus de détail, voir la Licence Publique Générale GNU .

Vous devez avoir reçu un exemplaire de la Licence Publique Générale GNU en même temps que ce programme ; si ce n'est pas le cas, écrivez à la Free Software Foundation Inc., 675 Mass Ave, Cambridge, MA 02139, Etats-Unis.
*/

package paellos.jdb;

import java.io.UnsupportedEncodingException;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import java.security.SecureRandom;
import java.sql.*;
import java.util.*;

public abstract class JDB<R extends Record> {
    protected String DBURL;
    protected String DBUSER;
    protected String DBPASSWORD;
    
    protected String tableName;
    protected String[] fields;
    protected String keyField;
    protected String createSQL;
    
    protected void init(final String DBURL, final String DBUSER, final String DBPASSWORD,
            final String tableName, final String[] fields, final String keyField, final String createSQL) {
        this.DBURL = DBURL;
        this.DBUSER = DBUSER;
        this.DBPASSWORD = DBPASSWORD;
        this.tableName = tableName;
        this.fields = fields;
        this.keyField = keyField;
        this.createSQL = createSQL;
    }
    
    public boolean tableExists() throws JDBException {
        Connection conn = openConnection();
        try {
            final DatabaseMetaData md = conn.getMetaData();
            final ResultSet rs = md.getTables(null, null, tableName, null);
            return(rs != null && rs.next());
        } catch (SQLException ex) {
            throw new JDBException("getTables failed: " + ex.getMessage());
        } finally {
            closeConnection(conn);
        }
    }
    
    protected Connection openConnection() throws JDBException {
        Connection conn;
        try {
            // Load the driver
            //Class.forName("org.gjt.mm.mysql.Driver");
            Class.forName("com.mysql.jdbc.Driver"); // nouvelle version
        /* or use the java option: -Djdbc.drivers=postgresql.Driver */

            // Connect to database
            final Properties connInfo = new Properties();
            
            connInfo.put("user",DBUSER);
            connInfo.put("password",DBPASSWORD);
            
            connInfo.put("useUnicode","true");
            connInfo.put("characterEncoding","UTF-8");
            
            connInfo.put("autoDeserialize","true"); // pour les Blobs
            
            conn = DriverManager.getConnection(DBURL, connInfo);

        } catch (ClassNotFoundException ex) {
            throw new JDBException("could not find the database driver class: "
                + ex.getMessage());
        } catch (SQLException ex) {
            throw new JDBException("getConnection failed: " + ex.getMessage());
        }
        
        return(conn);
    }
    
    protected static ResultSet query(final String q, final Statement st) throws JDBException {
        ResultSet rs;
        try {
            rs = st.executeQuery(q);
        } catch (SQLException ex) {
            throw new JDBException("executeQuery failed with the query '" +
                q + "': " + ex.getMessage());
        }
        return rs;
    }
    
    protected static Statement newStatement(final Connection conn) throws JDBException {
        Statement st;
        try {
            st = conn.createStatement();
        } catch (SQLException ex) {
            throw new JDBException("createStatement failed: " + ex.getMessage());
        }
        return(st);
    }
    
    protected static void closeStatement(final Statement st) throws JDBException {
        try {
            st.close();
        } catch (SQLException ex) {
            System.err.println("closeStatement failed: " + ex.getMessage());
        }
    }
    
    protected static void closeConnection(final Connection conn) throws JDBException {
        try {
            conn.close();
        } catch (SQLException ex) {
            throw new JDBException("Connection.close failed: " + ex.getMessage());
        }
    }
    
    protected static void closeNoError(final ResultSet rs, final Statement st) {
        if (rs != null) {
            try {
                rs.close();
            } catch (SQLException ex) {
            }
        }
        if (st != null) {
            try {
                st.close();
            } catch (SQLException ex) {
            }
        }
    }
    
    /**
     * Create a new empty database. Works only if it doesn't exist already.
     */
    public void createDB() throws JDBException {
        final Connection conn = openConnection();
        String sql = createSQL;
        Statement st = null;
        try {
            st = newStatement(conn);
            st.execute(sql);
        } catch (SQLException ex) {
            throw new JDBException("query '"+sql+"' failed: " + ex.getMessage());
        } finally {
            closeNoError(null, st);
            closeConnection(conn);
        }
    }
    
    public abstract R newRecord(final Object[] values) throws JDBException;
    
    /**
     * Add a new record to the database.
     * No check for existence, but the key value should be unique.
     */
    public void addRecord(final R rec) throws JDBException {
        final Connection conn = openConnection();
        final Object[] values = rec.getFieldValues();
        String sql = "insert into " + tableName + " values (";
        for (int i=0; i<values.length; i++) {
//            sql += "'" + values[i] + "'"; // pose des problèmes avec les '''
            sql += "?";
            if (i != fields.length-1)
                sql += ", ";
        }
        sql += ");";
        //Statement st = newStatement(conn);
        PreparedStatement st = null;
        try {
            try {
                st = conn.prepareStatement(sql);
                for (int i=0; i<values.length; i++)
                    st.setObject(i+1, values[i]);
            } catch (SQLException ex) {
                throw new JDBException("prepareStatement failed: " + ex.getMessage());
            }
            try {
                //st.execute(sql);
                st.execute();
            } catch (SQLException ex) {
                throw new JDBException("query '"+sql+"' failed: " + ex.getMessage());
            }
        } finally {
            closeNoError(null, st);
            closeConnection(conn);
        }
    }
    
    /**
     * Modify a record.
     * the key is used to specify the record, and the other fields are updated.
     */
    public void modRecord(final R rec) throws JDBException {
        final Connection conn = openConnection();
        final Object[] values = rec.getFieldValues();
        final Object keyValue = rec.getKeyValue();
        String sql = "update " + tableName + " set ";
        for (int i=0; i<fields.length; i++) {
            if (!keyField.equals(fields[i])) {
//                sql += fields[i] + " = '" + values[i] + "'"; // pose des problèmes avec les '''
                sql += fields[i] + " = ?";
                if (i != fields.length-1)
                    sql += ", ";
            }
        }
//        sql += " where " +keyField+ " = '" + keyValue + "';";
        sql += " where " +keyField+ " = ?;";
        //Statement st = newStatement(conn);
        PreparedStatement st = null;
        try {
            try {
                st = conn.prepareStatement(sql);
                int j=1;
                for (int i=0; i<fields.length; i++) {
                    if (!keyField.equals(fields[i])) {
                        st.setObject(j, values[i]);
                        j++;
                    }
                }
                st.setObject(j, keyValue);
            } catch (SQLException ex) {
                throw new JDBException("prepareStatement failed: " + ex.getMessage());
            }
            try {
                st.execute();
            } catch (SQLException ex) {
                throw new JDBException("query '"+sql+"' failed: " + ex.getMessage());
            }
        } finally {
            closeNoError(null, st);
            closeConnection(conn);
        }
    }
    
    /**
     * Delete a record.
     * the key is used to select the record, and the other fields are ignored.
     */
    public void delRecord(final R rec) throws JDBException {
        final Connection conn = openConnection();
        final Object keyValue = rec.getKeyValue();
        String sql = "delete from " + tableName + " where " +keyField+ " = '" + keyValue + "';";
        Statement st = null;
        try {
            st = newStatement(conn);
            st.execute(sql);
        } catch (SQLException ex) {
            throw new JDBException("query '"+sql+"' failed: " + ex.getMessage());
        } finally {
            closeNoError(null, st);
            closeConnection(conn);
        }
    }
    
    /**
     * Return a record from a key value
     */
    public R getRecord(final Object keyValue) throws JDBException {
        final Connection conn = openConnection();
        String sql = "select ";
        for (int i=0; i<fields.length; i++) {
            sql += fields[i];
            if (i != fields.length-1)
                sql += ", ";
        }
        sql += " from " + tableName + " where " +keyField+ " = '" + keyValue + "'";
        
        Statement st = null;
        ResultSet rs = null;
        R rec = null;
        try {
            st = newStatement(conn);
            rs = query(sql, st);
            if (rs == null)
                throw new JDBException("query failed: " + sql);
            try {
                if (rs.next()) {
                    final Object[] values = new Object[fields.length];
                    final ResultSetMetaData meta = rs.getMetaData();
                    for (int i=0; i<fields.length; i++) {
                        values[i] = rs.getObject(fields[i]);
                        if ("MEDIUMBLOB".equals(meta.getColumnTypeName(i+1)) && (values[i] != null) && !(values[i] instanceof Blob))
                            values[i] = rs.getBlob(fields[i]); // pour éviter un bug du driver mysql, qui renvoie un String
                    }
                    
                    rec = newRecord(values);
    
                } else {
                    throw new JDBException("no result for the query: " + sql);
                }
                if (rs.next())
                    System.err.println("Multiple records with key value " +keyValue.toString()+
                        " found in db");
            } catch (SQLException ex) {
                throw new JDBException("SQLException: " + ex.getMessage());
            }
        } finally {
            closeNoError(rs, st);
            closeConnection(conn);
        }
        return(rec);
    }
    
    /**
     * Return all records
     */
    public Vector<R> getAllRecords() throws JDBException {
        final Vector<R> records = new Vector<R>();
        final Connection conn = openConnection();
        String sql = "select ";
        for (int i=0; i<fields.length; i++) {
            sql += fields[i];
            if (i != fields.length-1)
                sql += ", ";
        }
        sql += " from " + tableName;
        Statement st = null;
        ResultSet rs = null;
        try {
            st = newStatement(conn);
            rs = query(sql, st);
            if (rs == null)
                throw new JDBException("query failed: " + sql);
            R rec = null;
            try {
                while (rs.next()) {
                    final Object[] values = new Object[fields.length];
                    final ResultSetMetaData meta = rs.getMetaData();
                    for (int i=0; i<fields.length; i++) {
                        values[i] = rs.getObject(fields[i]);
                        if ("MEDIUMBLOB".equals(meta.getColumnTypeName(i+1)) && (values[i] != null) && !(values[i] instanceof Blob))
                            values[i] = rs.getBlob(fields[i]); // pour éviter un bug du driver mysql, qui renvoie un String
                    }
                    
                    rec = newRecord(values);
                    records.add(rec);
                }
            } catch (SQLException ex) {
                throw new JDBException("SQLException: " + ex.getMessage());
            }
        } finally {
            closeNoError(rs, st);
            closeConnection(conn);
        }
        return(records);
    }
    
    /**
     * Find records with given values
     */
    public Vector<R> findRecords(final R rec) throws JDBException {
        final Vector<R> records = new Vector<R>();
        
        final Object[] values = rec.getFieldValues();
        String sql = "select ";
        for (int i=0; i<fields.length; i++) {
            sql += fields[i];
            if (i != fields.length-1)
                sql += ", ";
        }
        sql += " from " + tableName + " where ";
        boolean mod = false;
        for (int i=0; i<fields.length; i++) {
            if (values[i] != null) {
                if (!mod)
                    mod = true;
                else
                    sql += " and ";
                //sql += fields[i]+ " = '" + values[i] + "'"; // pose des problèmes avec les '''
                sql += fields[i] + " = ?";
            }
        }
        
        final Connection conn = openConnection();
        PreparedStatement st = null;
        ResultSet rs = null;
        try {
            try {
                st = conn.prepareStatement(sql);
                int j=1;
                for (int i=0; i<fields.length; i++) {
                    if (values[i] != null) {
                        st.setObject(j, values[i]);
                        j++;
                    }
                }
            } catch (SQLException ex) {
                throw new JDBException("prepareStatement failed: " + ex.getMessage());
            }
            
            try {
                rs = st.executeQuery();
            } catch (SQLException ex) {
                throw new JDBException("executeQuery failed with the query '" +
                    sql + "': " + ex.getMessage());
            }
            
            R rec2 = null;
            try {
                while (rs.next()) {
                    final ResultSetMetaData meta = rs.getMetaData();
                    for (int i=0; i<fields.length; i++) {
                        values[i] = rs.getObject(fields[i]);
                        if ("MEDIUMBLOB".equals(meta.getColumnTypeName(i+1)) && (values[i] != null) && !(values[i] instanceof Blob))
                            values[i] = rs.getBlob(fields[i]); // pour éviter un bug du driver mysql, qui renvoie un String
                    }
                    
                    rec2 = newRecord(values);
                    records.add(rec2);
                }
            } catch (SQLException ex) {
                throw new JDBException("SQLException: " + ex.getMessage());
            }
        } finally {
            closeNoError(rs, st);
            closeConnection(conn);
        }
        return(records);
    }

    /**
     * Returns the number of records in the table
     */
    public int nbRecords() throws JDBException {
        final Connection conn = openConnection();
        String sql = "select count(*) from " + tableName;
        
        Statement st = null;
        ResultSet rs = null;
        int nb = 0;
        try {
            st = newStatement(conn);
            rs = query(sql, st);
            if (rs == null)
                throw new JDBException("query failed: " + sql);
            try {
                if (rs.next())
                    nb = rs.getInt(1);
                else
                    throw new JDBException("no result for the query: " + sql);
            } catch (SQLException ex) {
                throw new JDBException("SQLException: " + ex.getMessage());
            }
        } finally {
            closeNoError(rs, st);
            closeConnection(conn);
        }
        return(nb);
    }
    
    /**
     * encrypt a password with the UNIX crypt function
     * (using the jcrypt library).
     */
    @Deprecated
    public static String encrypt(final String password) {
        final String letters = "0123456789abcdefghijklmnopqrstuvwxyz";
        final Random rand = new Random();
        final String salt =  "" + letters.charAt(rand.nextInt(letters.length())) +
            letters.charAt(rand.nextInt(letters.length()));
        return(jcrypt.crypt(salt, password));
    }
    
    /**
     * SHA-256 (32 bytes) with a 64 bits salt (8 bytes)
     * returns an 80 chars String
     */
    public static String encryptSHA256(final String password) throws NoSuchAlgorithmException {
        final MessageDigest digest = MessageDigest.getInstance("SHA-256");
        digest.reset();
        final byte[] bSalt = new byte[8];
        final SecureRandom random = new SecureRandom();
        random.nextBytes(bSalt);
        digest.update(bSalt);
        try {
            byte[] input = digest.digest(password.getBytes("UTF-8"));
            return(bytesToHex(bSalt) + bytesToHex(input));
        } catch (UnsupportedEncodingException ex) {
            return(null);
        }
    }
    
    /**
     * Test a password against an encrypted value (including the salt)
     */
    public static boolean testEnc(final String unenc, final String enc) throws NoSuchAlgorithmException {
        if (unenc == null || enc == null)
            return(false);
        if (enc.length() < 20) {
            // old method with crypt
            final String salt = enc.substring(0,2);
            final String unencenc = jcrypt.crypt(salt, unenc);
            return(unencenc.equals(enc));
        } else {
            // new method: SHA-256 (32 bytes) with a 64 bits salt (8 bytes)
            // requires 80 characters for hexa encoding
            // (base64 would use less bytes, but another field would be needed, or a special character to separate the salt
            // and it would require an additionnal library)
            final byte[] bSalt = hexToBytes(enc, 0, 16);
            final byte[] bRef = hexToBytes(enc, 16, 64);
            
            final MessageDigest digest = MessageDigest.getInstance("SHA-256");
            digest.reset();
            digest.update(bSalt);
            try {
                final byte[] bTest = digest.digest(unenc.getBytes("UTF-8"));
                return(Arrays.equals(bRef, bTest));
            } catch (UnsupportedEncodingException ex) {
                return(false);
            }
        }
    }
    
    private static String bytesToHex(final byte[] ba) {
        final StringBuffer hexString = new StringBuffer();
        for (int i=0; i<ba.length; i++) {
            final String hex = Integer.toHexString(0xFF & ba[i]);
            if (hex.length() == 1)
                hexString.append('0');
            hexString.append(hex);
        }
        return(hexString.toString());
    }
    
    private static byte[] hexToBytes(final String s, int offset, int length) {
        if ((length % 2) != 0)
            return null;
        final byte[] ba = new byte[length / 2];
        int j = 0;
        final int end = offset + length;
        for (int i = offset; i < end; i += 2) {
            final int high = Character.digit(s.charAt(i), 16);
            final int low = Character.digit(s.charAt(i+1), 16);
            if (high == -1 || low == -1)
                return null;
            ba[j++] = (byte)(((high << 4) & 0xF0) | (low & 0x0F));
        }
        return(ba);
    }
}
